package br.com.estruturadedados.utilitario.lista.encadeada.duplamente;

import br.com.estruturadedados.utilitario.lista.Lista;
import br.com.estruturadedados.utilitario.colecao.Iterator;

public class ListaDuplamenteEncadeada<E extends Comparable<E>> implements Lista<E> {

    private Celula<E> primeira;
    private Celula<E> ultima;
    private int tamanho;

    public ListaDuplamenteEncadeada() {
        this.primeira = null;
        this.ultima = null;
        this.tamanho = 0;
    }

    @Override
    public void adiciona(E elemento, int posicao) throws Exception {
        this.validarPosicao(posicao);
        this.verificarPosicaoOcupada(posicao);
        if (posicao == 0) {
            this.adicionaInicio(elemento);
        } else {
            if (posicao == this.tamanho) {
                this.adicionaFim(elemento);
            } else {
                Celula<E> anterior = this.recuperaCelulaRecursivo(this.primeira, posicao - 1, 0);
                Celula<E> proxima = anterior.getProxima();
                Celula<E> nova = new Celula<>(elemento, anterior.getProxima());
                nova.setAnterior(anterior);
                anterior.setProxima(nova);
                proxima.setAnterior(nova);
                this.tamanho += 1;
            }
        }
    }

    @Override
    public void adicionaInicio(E elemento) {
        if (this.vazio()) {
            Celula<E> nova = new Celula<>(elemento);
            this.primeira = nova;
            this.ultima = nova;
        } else {
            Celula<E> nova = new Celula<>(elemento, this.primeira);
            this.primeira.setAnterior(nova);
            this.primeira = nova;
        }
        this.tamanho += 1;
    }

    @Override
    public void adicionaFim(E elemento) {
        if (this.vazio()) {
            this.adicionaInicio(elemento);
        } else {
            Celula<E> nova = new Celula<>(elemento);
            this.ultima.setProxima(nova);
            nova.setAnterior(this.ultima);
            this.ultima = nova;
            this.tamanho += 1;
        }
    }

    @Override
    public E recupera(int posicao) throws Exception {
        this.validarPosicao(posicao);
        this.verificarSeEstaVazio();
        this.verificarPosicaoLivre(posicao);
        Celula<E> celulaRecuperada = this.recuperaCelulaRecursivo(this.primeira, posicao, 0);
        E elementoRecuperado = celulaRecuperada.getElemento();
        return elementoRecuperado;
    }

    @Override
    public boolean vazio() {
        return this.tamanho == 0;
    }

    @Override
    public void remove(int posicao) throws Exception {
        this.validarPosicao(posicao);
        this.verificarSeEstaVazio();
        this.verificarPosicaoLivre(posicao);
        if (posicao == 0) {
            this.removeInicio();
        } else {
            if (posicao == this.tamanho - 1) {
                this.removeFim();
            } else {
                Celula<E> anterior = this.recuperaCelulaRecursivo(this.primeira, posicao - 1, posicao);
                Celula<E> atual = anterior.getProxima();
                Celula<E> proxima = atual.getProxima();
                anterior.setProxima(proxima);
                proxima.setAnterior(anterior);
                this.tamanho -= 1;
            }
        }
    }

    @Override
    public void removeInicio() throws Exception {
        this.verificarSeEstaVazio();
        this.primeira = this.primeira.getProxima();
        this.tamanho -= 1;
        if (this.vazio()) {
            this.ultima = null;
        }
    }

    @Override
    public void removeFim() throws Exception {
        this.verificarSeEstaVazio();
        if (this.tamanho == 1) {
            this.removeInicio();
        } else {
            Celula<E> penultima = this.ultima.getAnterior();
            penultima.setProxima(null);
            this.ultima = penultima;
            this.tamanho -= 1;
        }
    }

    @Override
    public int tamanho() {
        return this.tamanho;
    }

    @Override
    public void limpa() throws Exception {
        this.verificarSeEstaVazio();
        this.primeira = null;
        this.ultima = null;
    }

    @Override
    public Iterator<E> iterator() {
        Iterator<E> iterator = new Iterador<>(this.primeira, this.ultima);
        return iterator;
    }

    private Celula<E> recuperaCelulaIteraivo(int posicao) throws Exception {
        this.validarPosicao(posicao);
        this.verificarSeEstaVazio();
        this.verificarPosicaoLivre(posicao);
        Celula atual = this.primeira;
        for (int i = 0; i < posicao; i++) {
            atual = atual.getProxima();
        }
        return atual;
    }

    private Celula<E> recuperaCelulaRecursivo(Celula<E> celula, int posicao, int posicaoAtual) throws Exception {
        this.validarPosicao(posicao);
        this.verificarSeEstaVazio();
        this.verificarPosicaoLivre(posicao);
        return celula == null ? null : posicao == 0 ? celula : posicaoAtual < posicao ? this.recuperaCelulaRecursivo(celula.getProxima(), posicao, posicaoAtual + 1) : celula;
    }

    private void validarPosicao(int posicao) throws Exception {
        if (posicao < 0 && posicao >= this.tamanho) {
            Exception exception = new Exception("Posição inválida.");
            throw exception;
        }
    }

    private void verificarPosicaoOcupada(int posicao) throws Exception {
        if (posicao >= 0 && posicao < this.tamanho) {
            Exception exception = new Exception("Posição ocupada.");
            throw exception;
        }
    }

    private void verificarPosicaoLivre(int posicao) throws Exception {
        if (posicao < 0 && posicao > this.tamanho) {
            Exception exception = new Exception("Posição desoculpada.");
            throw exception;
        }
    }

    private void verificarSeEstaVazio() throws Exception {
        if (this.vazio()) {
            Exception exception = new Exception("Está vazio.");
            throw exception;
        }
    }

    public void imprimeIterativo(char tipo) {
        if (tipo == 'C' || tipo == 'c') {
            for (Iterator iterator = this.iterator(); iterator.hasNext();) {
                System.out.println(iterator.next());
            }
        }
        if (tipo == 'D' || tipo == 'd') {
            for (Iterator iterator = this.iterator(); iterator.hasPrevious();) {
                System.out.println(iterator.previous());
            }
        }
    }

    public void imprimeRecursivo(char tipo) {
        if (tipo == 'C' || tipo == 'c') {
            this.crescente(this.primeira);
        }
        if (tipo == 'D' || tipo == 'd') {
            this.descrescente(this.primeira);
        }
    }

    private void crescente(Celula<E> celula) {
        if (celula != null) {
            System.out.println(celula.getElemento());
            this.crescente(celula.getProxima());
        }
    }

    private void descrescente(Celula<E> celula) {
        if (celula != null) {
            this.descrescente(celula.getProxima());
            System.out.println(celula.getElemento());
        }
    }

    public Celula<E> getPrimeira() {
        return primeira;
    }

    public void setPrimeira(Celula<E> primeira) {
        this.primeira = primeira;
    }

    public Celula<E> getUltima() {
        return ultima;
    }

    public void setUltima(Celula<E> ultima) {
        this.ultima = ultima;
    }

}
