package br.com.estruturadedados.utilitario.lista.encadeada.simplismente;

import br.com.estruturadedados.utilitario.colecao.Iterator;
import br.com.estruturadedados.utilitario.lista.Lista;

public class ListaSimplismenteEncadeada<E extends Comparable<E>> implements Lista<E> {

    private Celula<E> primeira;
    private Celula<E> ultima;
    private int tamanho;

    public ListaSimplismenteEncadeada() {
        this.primeira = null;
        this.ultima = null;
        this.tamanho = 0;
    }

    @Override
    public void adiciona(E elemento, int posicao) throws Exception {
        this.validarPosicao(posicao);
        this.verificarPosicaoOcupada(posicao);
        if (posicao == 0) {
            this.adicionaInicio(elemento);
        } else {
            if (posicao == this.tamanho) {
                this.adicionaFim(elemento);
            } else {
                Celula<E> anterior = this.recuperaCelulaRecursivo((Celula<E>) this.primeira, posicao - 1, 0);
                Celula<E> nova = new Celula<>(anterior.getProxima(), elemento);
                anterior.setProxima(nova);
                this.tamanho += 1;
            }
        }
    }

    @Override
    public void adicionaInicio(E elemento) {
        Celula<E> nova = new Celula<>((Celula<E>) this.primeira, elemento);
        this.primeira = nova;
        if (this.vazio()) {
            this.ultima = this.primeira;
        }
        this.tamanho += 1;
    }

    @Override
    public void adicionaFim(E elemento) {
        if (this.vazio()) {
            this.adicionaInicio(elemento);
        } else {
            Celula<E> nova = new Celula<>(elemento);
            this.ultima.setProxima(nova);
            this.ultima = nova;
            this.tamanho += 1;
        }
    }

    @Override
    public E recupera(int posicao) throws Exception {
        this.validarPosicao(posicao);
        this.verificarSeEstaVazio();
        this.verificarPosicaoLivre(posicao);
        Celula<E> celulaRecuperada = this.recuperaCelulaRecursivo((Celula<E>) this.primeira, posicao, 0);
        E elementoRecuperado = (E) celulaRecuperada.getElemento();
        return elementoRecuperado;
    }

    @Override
    public boolean vazio() {
        return this.tamanho == 0;
    }

    @Override
    public void remove(int posicao) throws Exception {
        this.validarPosicao(posicao);
        this.verificarSeEstaVazio();
        this.verificarPosicaoLivre(posicao);
        if (this.vazio()) {
            this.removeInicio();
        } else {
            if (posicao == this.tamanho - 1) {
                this.removeFim();
            } else {
                Celula<E> anterior = this.recuperaCelulaRecursivo((Celula<E>) this.primeira, posicao - 1, 0);
                Celula<E> atual = anterior.getProxima();
                Celula<E> proxima = atual.getProxima();
                anterior.setProxima(proxima);
                proxima.setElemento(anterior.getProxima().getElemento());
                this.tamanho -= 1;
            }
        }
    }

    @Override
    public void removeInicio() throws Exception {
        this.verificarSeEstaVazio();
        this.primeira = this.primeira.getProxima();
        this.tamanho -= 1;
        if (this.vazio()) {
            this.ultima = null;
        }
    }

    @Override
    public void removeFim() throws Exception {
        this.verificarSeEstaVazio();
        Celula<E> celulaAuxiliar = (Celula<E>) this.primeira;
        Celula<E> celulaAnterior = null;
        while (celulaAuxiliar.getProxima() != null) {
            celulaAnterior = celulaAuxiliar;
            celulaAuxiliar = celulaAuxiliar.getProxima();
        }
        celulaAnterior.setProxima(null);
        this.ultima = celulaAnterior;
        this.tamanho -= 1;
    }

    @Override
    public int tamanho() {
        return this.tamanho;
    }

    @Override
    public void limpa() throws Exception {
        this.verificarSeEstaVazio();
        this.primeira = null;
        this.ultima = null;
    }

    @Override
    public Iterator<E> iterator() {
        Iterator<E> iterator = new Iterador<>((Celula<E>) this.primeira, (Celula<E>) this.ultima);
        return iterator;
    }

    private Celula<E> recuperaCelulaIterativa(int posicao) throws Exception {
        this.validarPosicao(posicao);
        this.verificarSeEstaVazio();
        this.verificarPosicaoLivre(posicao);
        Celula atual = this.primeira;
        for (int i = 0; i < posicao; i++) {
            atual = atual.getProxima();
        }
        return atual;
    }

    private Celula<E> recuperaCelulaRecursivo(Celula<E> celula, int posicao, int posicaoAtual) throws Exception {
        this.validarPosicao(posicao);
        this.verificarSeEstaVazio();
        this.verificarPosicaoLivre(posicao);
        return celula == null ? null : posicao == 0 ? celula : posicaoAtual < posicao ? this.recuperaCelulaRecursivo(celula.getProxima(), posicao, posicaoAtual + 1) : celula;
    }

    private void validarPosicao(int posicao) throws Exception {
        if (posicao < 0 && posicao >= this.tamanho) {
            Exception exception = new Exception("Posição inválida.");
            throw exception;
        }
    }

    private void verificarPosicaoOcupada(int posicao) throws Exception {
        if (posicao >= 0 && posicao < this.tamanho) {
            Exception exception = new Exception("Posição ocupada.");
            throw exception;
        }
    }

    private void verificarPosicaoLivre(int posicao) throws Exception {
        if (!(posicao >= 0 && posicao < this.tamanho)) {
            Exception exception = new Exception("Posição desoculpada.");
            throw exception;
        }
    }

    private void verificarSeEstaVazio() throws Exception {
        if (this.vazio()) {
            Exception exception = new Exception("Está vazio.");
            throw exception;
        }
    }

    public Celula<E> getPrimeira() {
        return primeira;
    }

    public void setPrimeira(Celula<E> primeira) {
        this.primeira = primeira;
    }

    public Celula<E> getUltima() {
        return ultima;
    }

    public void setUltima(Celula<E> ultima) {
        this.ultima = ultima;
    }

}
